#!env python3

import argparse
import easygraph as eg
import enum
import functools
import glob
import graphviz
import io
import itertools
import math
import numpy
import os
import platform
import random
import re
import scipy
import subprocess
import sys
import tempfile
import warnings


QUERY_FILE_PATTERN = re.compile(r'\d+\w\.sql')
JOIN_PATTERN = re.compile(r'(\w+)\.\w+ = (\w+)\.\w+')


#=======================================================================================================================
# Helper functions
#=======================================================================================================================

def default_open(filepath):
    if platform.system() == 'Darwin':       # macOS
        subprocess.call(('open', filepath))
    elif platform.system() == 'Windows':    # Windows
        os.startfile(filepath)
    else:                                   # linux variants
        subprocess.call(('xdg-open', filepath))

def compute_graph_density(G :eg.Graph):
    n = G.number_of_nodes()
    m = G.number_of_edges()
    return 2 * m / (n * (n-1))

def compute_edge_skewness(G :eg.Graph):
    degrees = sorted(list(G.degree().values()))
    try:
        with warnings.catch_warnings():
            warnings.simplefilter('error') # turn warnings into errors
            return abs(scipy.stats.skew(degrees, nan_policy='raise')) # abs(): we don't care whether tail is left or right
    except RuntimeWarning:
        return 0 # on RuntimeWarning, the degrees are identical ⇒ return skew 0

def compute_edge_p_value(G :eg.Graph):
    degrees = list(G.degree().values())
    _, p = scipy.stats.chisquare(f_obs=degrees)
    return p

def compute_edge_stddev(G :eg.Graph):
    degrees = list(G.degree().values())
    return scipy.stats.tstd(degrees)

def compute_edge_entropy(G :eg.Graph):
    degrees = list(G.degree().values())
    mean = len(G.edges) // len(G.nodes)
    avg_degrees = [mean] * len(G.nodes)
    for i in range(len(G.edges) - mean * len(G.nodes)):
        avg_degrees[i] += 1
    return scipy.stats.entropy(pk=degrees, qk=avg_degrees)

def show_graph(filename :str, G :eg.Graph):
    with tempfile.NamedTemporaryFile(mode='w+', encoding='utf-8') as dotfile:
        eg.write_dot(G, dotfile.name)
        dotfile.seek(0)
        #  with tempfile.NamedTemporaryFile() as pdf:
        with open(f'{filename}.pdf', 'wb') as pdf:
            source = dotfile.read()
            pdf.write(
                graphviz.Source(source, format='pdf', engine='circo').pipe()
            )
            pdf.flush()
            default_open(pdf.name)


#=======================================================================================================================
# Graph construction
#=======================================================================================================================

def query_graph_from_file(file: io.TextIOWrapper) -> eg.Graph:
    G = eg.Graph()

    #===== Process joins ====
    for line in file.readlines():
        for join in JOIN_PATTERN.findall(line):
            lhs, rhs = join
            G.add_edge(lhs, rhs)

    assert eg.is_connected(G), 'Query graph must be connected'
    return G


#=======================================================================================================================
# main
#=======================================================================================================================

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Compute information of JOB queries.')
    parser.add_argument('JOB', help='Path to the join order benchmark (JOB)', type=str, action='store',
                        metavar='JOB')
    parser.add_argument('--show', help='Directly show the graph of the generated query', action='store_true')
    args = parser.parse_args()

    print('query,vertices,edges,density,skewness,p_value,stddev,entropy') # header
    directory = os.fsencode(args.JOB)
    for file in os.listdir(directory):
        filename = os.fsdecode(file)
        if QUERY_FILE_PATTERN.match(filename):
            with open(os.path.join(args.JOB, filename), 'r') as query_file:
                G = query_graph_from_file(query_file)
                basename = os.path.splitext(filename)[0]
                density = compute_graph_density(G)
                skewness = compute_edge_skewness(G)
                p_value = compute_edge_p_value(G)
                stddev = compute_edge_stddev(G)
                entropy = compute_edge_entropy(G)
                print(f'{basename},{len(G.nodes)},{len(G.edges)},{density:.3f},{skewness:.3f},{p_value:.3f},{stddev:.3f},{entropy:.3f}')
                if args.show:
                    show_graph(basename, G)
